
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#undef GTK_DISABLE_DEPRECATED
#define GTK_ENABLE_BROKEN
#include <dialogs/mime-settings/xfce-mime-helper.h>

#include "xfce-mime-helper-enum-types.h"

/* enumerations from "xfce-mime-helper.h" */
GType
xfce_mime_helper_category_get_type (void)
{
	static GType type = 0;
	if (type == 0) {
	static const GEnumValue values[] = {
	{ XFCE_MIME_HELPER_WEBBROWSER, "XFCE_MIME_HELPER_WEBBROWSER", "WebBrowser" },
	{ XFCE_MIME_HELPER_MAILREADER, "XFCE_MIME_HELPER_MAILREADER", "MailReader" },
	{ XFCE_MIME_HELPER_FILEMANAGER, "XFCE_MIME_HELPER_FILEMANAGER", "FileManager" },
	{ XFCE_MIME_HELPER_TERMINALEMULATOR, "XFCE_MIME_HELPER_TERMINALEMULATOR", "TerminalEmulator" },
	{ 0, NULL, NULL }
	};
	type = g_enum_register_static ("XfceMimeHelperCategory", values);
  }
	return type;
}

/* Generated data ends here */

