
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef __XFCE_MIME_HELPER_ENUM_TYPES_H__
#define __XFCE_MIME_HELPER_ENUM_TYPES_H__
#include <exo/exo.h>
G_BEGIN_DECLS
/* enumerations from "xfce-mime-helper.h" */
GType xfce_mime_helper_category_get_type (void) G_GNUC_CONST;
#define XFCE_MIME_TYPE_MIME_HELPER_CATEGORY (xfce_mime_helper_category_get_type())
G_END_DECLS

#endif /* !__XFCE_MIME_HELPER_ENUM_TYPES_H__ */

/* Generated data ends here */

