#if G_GNUC_CHECK_VERSION (4, 6)
# pragma GCC diagnostic push
# pragma GCC diagnostic ignored "-Wunused-variable"
#endif
static const gchar *xsettings_properties_Net[] = {
 "ThemeName",
 "IconThemeName",
 "DoubleClickTime",
 "DoubleClickDistance",
 "DndDragThreshold",
 "CursorBlink",
 "CursorBlinkTime",
 "SoundThemeName",
 "EnableEventSounds",
 "EnableInputFeedbackSounds",
};
static const gchar *xsettings_properties_Xft[] = {
 "DPI",
 "Antialias",
 "Hinting",
 "HintStyle",
 "RGBA",
};
static const gchar *xsettings_properties_Gtk[] = {
 "CanChangeAccels",
 "ColorPalette",
 "FontName",
 "MonospaceFontName",
 "IconSizes",
 "KeyThemeName",
 "MenuImages",
 "ButtonImages",
 "MenuBarAccel",
 "CursorThemeName",
 "CursorThemeSize",
 "DecorationLayout",
 "DialogsUseHeader",
 "TitlebarMiddleClick",
};
static const gchar *xsettings_properties_Gdk[] = {
 "WindowScalingFactor",
};
#if G_GNUC_CHECK_VERSION (4, 6)
# pragma GCC diagnostic pop
#endif
