/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __GTK_SETTINGS_EXPORTED_H__
#define __GTK_SETTINGS_EXPORTED_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gtk.Settings */

#define XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED (xfce_gtk_settings_helper_exported_get_type ())
#define XFCE_GTK_SETTINGS_HELPER_EXPORTED(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED, XfceGtkSettingsHelperExported))
#define XFCE_GTK_SETTINGS_HELPER_IS_EXPORTED(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED))
#define XFCE_GTK_SETTINGS_HELPER_EXPORTED_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED, XfceGtkSettingsHelperExportedIface))

struct _XfceGtkSettingsHelperExported;
typedef struct _XfceGtkSettingsHelperExported XfceGtkSettingsHelperExported;
typedef struct _XfceGtkSettingsHelperExportedIface XfceGtkSettingsHelperExportedIface;

struct _XfceGtkSettingsHelperExportedIface
{
  GTypeInterface parent_iface;

  const gchar * (*get_modules) (XfceGtkSettingsHelperExported *object);

};

GType xfce_gtk_settings_helper_exported_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *xfce_gtk_settings_helper_exported_interface_info (void);
guint xfce_gtk_settings_helper_exported_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus property accessors: */
const gchar *xfce_gtk_settings_helper_exported_get_modules (XfceGtkSettingsHelperExported *object);
gchar *xfce_gtk_settings_helper_exported_dup_modules (XfceGtkSettingsHelperExported *object);
void xfce_gtk_settings_helper_exported_set_modules (XfceGtkSettingsHelperExported *object, const gchar *value);


/* ---- */

#define XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_PROXY (xfce_gtk_settings_helper_exported_proxy_get_type ())
#define XFCE_GTK_SETTINGS_HELPER_EXPORTED_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_PROXY, XfceGtkSettingsHelperExportedProxy))
#define XFCE_GTK_SETTINGS_HELPER_EXPORTED_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_PROXY, XfceGtkSettingsHelperExportedProxyClass))
#define XFCE_GTK_SETTINGS_HELPER_EXPORTED_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_PROXY, XfceGtkSettingsHelperExportedProxyClass))
#define XFCE_GTK_SETTINGS_HELPER_IS_EXPORTED_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_PROXY))
#define XFCE_GTK_SETTINGS_HELPER_IS_EXPORTED_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_PROXY))

typedef struct _XfceGtkSettingsHelperExportedProxy XfceGtkSettingsHelperExportedProxy;
typedef struct _XfceGtkSettingsHelperExportedProxyClass XfceGtkSettingsHelperExportedProxyClass;
typedef struct _XfceGtkSettingsHelperExportedProxyPrivate XfceGtkSettingsHelperExportedProxyPrivate;

struct _XfceGtkSettingsHelperExportedProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XfceGtkSettingsHelperExportedProxyPrivate *priv;
};

struct _XfceGtkSettingsHelperExportedProxyClass
{
  GDBusProxyClass parent_class;
};

GType xfce_gtk_settings_helper_exported_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XfceGtkSettingsHelperExportedProxy, g_object_unref)
#endif

void xfce_gtk_settings_helper_exported_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XfceGtkSettingsHelperExported *xfce_gtk_settings_helper_exported_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XfceGtkSettingsHelperExported *xfce_gtk_settings_helper_exported_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xfce_gtk_settings_helper_exported_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XfceGtkSettingsHelperExported *xfce_gtk_settings_helper_exported_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XfceGtkSettingsHelperExported *xfce_gtk_settings_helper_exported_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_SKELETON (xfce_gtk_settings_helper_exported_skeleton_get_type ())
#define XFCE_GTK_SETTINGS_HELPER_EXPORTED_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_SKELETON, XfceGtkSettingsHelperExportedSkeleton))
#define XFCE_GTK_SETTINGS_HELPER_EXPORTED_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_SKELETON, XfceGtkSettingsHelperExportedSkeletonClass))
#define XFCE_GTK_SETTINGS_HELPER_EXPORTED_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_SKELETON, XfceGtkSettingsHelperExportedSkeletonClass))
#define XFCE_GTK_SETTINGS_HELPER_IS_EXPORTED_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_SKELETON))
#define XFCE_GTK_SETTINGS_HELPER_IS_EXPORTED_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XFCE_GTK_SETTINGS_HELPER_TYPE_EXPORTED_SKELETON))

typedef struct _XfceGtkSettingsHelperExportedSkeleton XfceGtkSettingsHelperExportedSkeleton;
typedef struct _XfceGtkSettingsHelperExportedSkeletonClass XfceGtkSettingsHelperExportedSkeletonClass;
typedef struct _XfceGtkSettingsHelperExportedSkeletonPrivate XfceGtkSettingsHelperExportedSkeletonPrivate;

struct _XfceGtkSettingsHelperExportedSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XfceGtkSettingsHelperExportedSkeletonPrivate *priv;
};

struct _XfceGtkSettingsHelperExportedSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xfce_gtk_settings_helper_exported_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XfceGtkSettingsHelperExportedSkeleton, g_object_unref)
#endif

XfceGtkSettingsHelperExported *xfce_gtk_settings_helper_exported_skeleton_new (void);


G_END_DECLS

#endif /* __GTK_SETTINGS_EXPORTED_H__ */
