
#include "foo-marshal.h"


#ifdef G_ENABLE_DEBUG
#define g_marshal_value_peek_boolean(v) g_value_get_boolean (v)
#define g_marshal_value_peek_char(v) g_value_get_char (v)
#define g_marshal_value_peek_uchar(v) g_value_get_uchar (v)
#define g_marshal_value_peek_int(v) g_value_get_int (v)
#define g_marshal_value_peek_uint(v) g_value_get_uint (v)
#define g_marshal_value_peek_long(v) g_value_get_long (v)
#define g_marshal_value_peek_ulong(v) g_value_get_ulong (v)
#define g_marshal_value_peek_int64(v) g_value_get_int64 (v)
#define g_marshal_value_peek_uint64(v) g_value_get_uint64 (v)
#define g_marshal_value_peek_enum(v) g_value_get_enum (v)
#define g_marshal_value_peek_flags(v) g_value_get_flags (v)
#define g_marshal_value_peek_float(v) g_value_get_float (v)
#define g_marshal_value_peek_double(v) g_value_get_double (v)
#define g_marshal_value_peek_string(v) (char *) g_value_get_string (v)
#define g_marshal_value_peek_param(v) g_value_get_param (v)
#define g_marshal_value_peek_boxed(v) g_value_get_boxed (v)
#define g_marshal_value_peek_pointer(v) g_value_get_pointer (v)
#define g_marshal_value_peek_object(v) g_value_get_object (v)
#else /* !G_ENABLE_DEBUG */
/* WARNING: This code accesses GValues directly, which is UNSUPPORTED API.
 *          Do not access GValues directly in your code. Instead, use the
 *          g_value_get_*() functions
 */
#define g_marshal_value_peek_boolean(v) (v)->data[0].v_int
#define g_marshal_value_peek_char(v) (v)->data[0].v_int
#define g_marshal_value_peek_uchar(v) (v)->data[0].v_uint
#define g_marshal_value_peek_int(v) (v)->data[0].v_int
#define g_marshal_value_peek_uint(v) (v)->data[0].v_uint
#define g_marshal_value_peek_long(v) (v)->data[0].v_long
#define g_marshal_value_peek_ulong(v) (v)->data[0].v_ulong
#define g_marshal_value_peek_int64(v) (v)->data[0].v_int64
#define g_marshal_value_peek_uint64(v) (v)->data[0].v_uint64
#define g_marshal_value_peek_enum(v) (v)->data[0].v_long
#define g_marshal_value_peek_flags(v) (v)->data[0].v_ulong
#define g_marshal_value_peek_float(v) (v)->data[0].v_float
#define g_marshal_value_peek_double(v) (v)->data[0].v_double
#define g_marshal_value_peek_string(v) (v)->data[0].v_pointer
#define g_marshal_value_peek_param(v) (v)->data[0].v_pointer
#define g_marshal_value_peek_boxed(v) (v)->data[0].v_pointer
#define g_marshal_value_peek_pointer(v) (v)->data[0].v_pointer
#define g_marshal_value_peek_object(v) (v)->data[0].v_pointer
#endif /* !G_ENABLE_DEBUG */


/* VOID:OBJECT,OBJECT (marshal.list:1) */
void
foo_marshal_VOID__OBJECT_OBJECT (GClosure *closure,
                                 GValue *return_value G_GNUC_UNUSED,
                                 guint n_param_values,
                                 const GValue *param_values,
                                 gpointer invocation_hint G_GNUC_UNUSED,
                                 gpointer marshal_data)
{
    typedef void (*GMarshalFunc_VOID__OBJECT_OBJECT) (gpointer data1,
                                                      gpointer arg_1,
                                                      gpointer arg_2,
                                                      gpointer data2);
    register GMarshalFunc_VOID__OBJECT_OBJECT callback;
    register GCClosure *cc = (GCClosure *) closure;
    register gpointer data1, data2;

    g_return_if_fail (n_param_values == 3);

    if (G_CCLOSURE_SWAP_DATA (closure))
    {
        data1 = closure->data;
        data2 = g_value_peek_pointer (param_values + 0);
    }
    else
    {
        data1 = g_value_peek_pointer (param_values + 0);
        data2 = closure->data;
    }
    callback = (GMarshalFunc_VOID__OBJECT_OBJECT) (marshal_data ? marshal_data : cc->callback);

    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_object (param_values + 2),
              data2);
}

/* VOID:UINT,UINT,UINT,UINT (marshal.list:2) */
void
foo_marshal_VOID__UINT_UINT_UINT_UINT (GClosure *closure,
                                       GValue *return_value G_GNUC_UNUSED,
                                       guint n_param_values,
                                       const GValue *param_values,
                                       gpointer invocation_hint G_GNUC_UNUSED,
                                       gpointer marshal_data)
{
    typedef void (*GMarshalFunc_VOID__UINT_UINT_UINT_UINT) (gpointer data1,
                                                            guint arg_1,
                                                            guint arg_2,
                                                            guint arg_3,
                                                            guint arg_4,
                                                            gpointer data2);
    register GMarshalFunc_VOID__UINT_UINT_UINT_UINT callback;
    register GCClosure *cc = (GCClosure *) closure;
    register gpointer data1, data2;

    g_return_if_fail (n_param_values == 5);

    if (G_CCLOSURE_SWAP_DATA (closure))
    {
        data1 = closure->data;
        data2 = g_value_peek_pointer (param_values + 0);
    }
    else
    {
        data1 = g_value_peek_pointer (param_values + 0);
        data2 = closure->data;
    }
    callback = (GMarshalFunc_VOID__UINT_UINT_UINT_UINT) (marshal_data ? marshal_data : cc->callback);

    callback (data1,
              g_marshal_value_peek_uint (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_uint (param_values + 3),
              g_marshal_value_peek_uint (param_values + 4),
              data2);
}

/* VOID:UINT,UINT (marshal.list:3) */
void
foo_marshal_VOID__UINT_UINT (GClosure *closure,
                             GValue *return_value G_GNUC_UNUSED,
                             guint n_param_values,
                             const GValue *param_values,
                             gpointer invocation_hint G_GNUC_UNUSED,
                             gpointer marshal_data)
{
    typedef void (*GMarshalFunc_VOID__UINT_UINT) (gpointer data1,
                                                  guint arg_1,
                                                  guint arg_2,
                                                  gpointer data2);
    register GMarshalFunc_VOID__UINT_UINT callback;
    register GCClosure *cc = (GCClosure *) closure;
    register gpointer data1, data2;

    g_return_if_fail (n_param_values == 3);

    if (G_CCLOSURE_SWAP_DATA (closure))
    {
        data1 = closure->data;
        data2 = g_value_peek_pointer (param_values + 0);
    }
    else
    {
        data1 = g_value_peek_pointer (param_values + 0);
        data2 = closure->data;
    }
    callback = (GMarshalFunc_VOID__UINT_UINT) (marshal_data ? marshal_data : cc->callback);

    callback (data1,
              g_marshal_value_peek_uint (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              data2);
}

/* VOID:BOXED (marshal.list:4) */

/* VOID:BOXED,BOXED (marshal.list:5) */
void
foo_marshal_VOID__BOXED_BOXED (GClosure *closure,
                               GValue *return_value G_GNUC_UNUSED,
                               guint n_param_values,
                               const GValue *param_values,
                               gpointer invocation_hint G_GNUC_UNUSED,
                               gpointer marshal_data)
{
    typedef void (*GMarshalFunc_VOID__BOXED_BOXED) (gpointer data1,
                                                    gpointer arg_1,
                                                    gpointer arg_2,
                                                    gpointer data2);
    register GMarshalFunc_VOID__BOXED_BOXED callback;
    register GCClosure *cc = (GCClosure *) closure;
    register gpointer data1, data2;

    g_return_if_fail (n_param_values == 3);

    if (G_CCLOSURE_SWAP_DATA (closure))
    {
        data1 = closure->data;
        data2 = g_value_peek_pointer (param_values + 0);
    }
    else
    {
        data1 = g_value_peek_pointer (param_values + 0);
        data2 = closure->data;
    }
    callback = (GMarshalFunc_VOID__BOXED_BOXED) (marshal_data ? marshal_data : cc->callback);

    callback (data1,
              g_marshal_value_peek_boxed (param_values + 1),
              g_marshal_value_peek_boxed (param_values + 2),
              data2);
}

/* VOID:POINTER,BOXED,POINTER (marshal.list:6) */
void
foo_marshal_VOID__POINTER_BOXED_POINTER (GClosure *closure,
                                         GValue *return_value G_GNUC_UNUSED,
                                         guint n_param_values,
                                         const GValue *param_values,
                                         gpointer invocation_hint G_GNUC_UNUSED,
                                         gpointer marshal_data)
{
    typedef void (*GMarshalFunc_VOID__POINTER_BOXED_POINTER) (gpointer data1,
                                                              gpointer arg_1,
                                                              gpointer arg_2,
                                                              gpointer arg_3,
                                                              gpointer data2);
    register GMarshalFunc_VOID__POINTER_BOXED_POINTER callback;
    register GCClosure *cc = (GCClosure *) closure;
    register gpointer data1, data2;

    g_return_if_fail (n_param_values == 4);

    if (G_CCLOSURE_SWAP_DATA (closure))
    {
        data1 = closure->data;
        data2 = g_value_peek_pointer (param_values + 0);
    }
    else
    {
        data1 = g_value_peek_pointer (param_values + 0);
        data2 = closure->data;
    }
    callback = (GMarshalFunc_VOID__POINTER_BOXED_POINTER) (marshal_data ? marshal_data : cc->callback);

    callback (data1,
              g_marshal_value_peek_pointer (param_values + 1),
              g_marshal_value_peek_boxed (param_values + 2),
              g_marshal_value_peek_pointer (param_values + 3),
              data2);
}

/* VOID:POINTER,POINTER (marshal.list:7) */
void
foo_marshal_VOID__POINTER_POINTER (GClosure *closure,
                                   GValue *return_value G_GNUC_UNUSED,
                                   guint n_param_values,
                                   const GValue *param_values,
                                   gpointer invocation_hint G_GNUC_UNUSED,
                                   gpointer marshal_data)
{
    typedef void (*GMarshalFunc_VOID__POINTER_POINTER) (gpointer data1,
                                                        gpointer arg_1,
                                                        gpointer arg_2,
                                                        gpointer data2);
    register GMarshalFunc_VOID__POINTER_POINTER callback;
    register GCClosure *cc = (GCClosure *) closure;
    register gpointer data1, data2;

    g_return_if_fail (n_param_values == 3);

    if (G_CCLOSURE_SWAP_DATA (closure))
    {
        data1 = closure->data;
        data2 = g_value_peek_pointer (param_values + 0);
    }
    else
    {
        data1 = g_value_peek_pointer (param_values + 0);
        data2 = closure->data;
    }
    callback = (GMarshalFunc_VOID__POINTER_POINTER) (marshal_data ? marshal_data : cc->callback);

    callback (data1,
              g_marshal_value_peek_pointer (param_values + 1),
              g_marshal_value_peek_pointer (param_values + 2),
              data2);
}
